from __future__ import annotations
from typing import TypeVar
import numpy as np

from m3d.vector import Vector
from m3d.orientation import Orientation
from m3d.common import NumberType, float_eps


T = TypeVar("T")


class Transform:
    """
    Create a new Transform object
    Accepts an orientation and a vector or a matrix 4*4 as argument
    Rmq:
    When creating a transform from a 4*4 Matrix, the matrix is directly used
    as the Transform data
    When accessing/modifying the Orientation or Vector object you are
    modifying a vew of the matrix data
    When creating a new Transform object from an Orientation and
    Vector or 2 numpy arrays, you are copying them
    """

    def __init__(self, orientation=None, vector=None, matrix=None, dtype=np.float32, frozen=False) -> None:
        if matrix is not None:
            self._data = matrix
        else:
            self._data = np.identity(4, dtype=dtype)
        if orientation is None:
            pass
        elif isinstance(orientation, np.ndarray):
            if orientation.shape == (3, 3):
                self._data[:3, :3] = orientation
            else:
                raise ValueError()
        elif isinstance(orientation, Orientation):
            self._data[:3, :3] = orientation.data
        else:
            raise ValueError("orientation argument should be a numpy array, Orientation or None")

        if vector is None:
            pass
        elif isinstance(vector, np.ndarray):
            self._data[:3, 3] = vector
        elif isinstance(vector, Vector):
            self._data[:3, 3] = vector.data
        elif isinstance(vector, (list, tuple)):
            self._data[:3, 3] = vector
        else:
            raise ValueError("orientation argument should be a numpy array, Orientation or None")
        self._frozen = frozen
        self._data.flags.writeable = not frozen
        self.dtype = dtype

    def is_valid(self) -> bool:
        """
        Check if a transform is valid
        """
        if abs(self._data[3, 3] - 1) > float_eps:
            return False
        if not (abs(self._data[3, 0:3]) < float_eps).all():
            return False
        if np.isnan(self._data.sum()):
            return False
        return self.orient.is_valid()

    def _validate(self) -> Transform:
        if not self.is_valid():
            raise ValueError("Given args gives an invalid Transform")
        return self

    @property
    def frozen(self) -> bool:
        return self._frozen

    @frozen.setter
    def frozen(self, val: bool) -> None:
        self._frozen = val
        self._data.flags.writeable = not val

    def __str__(self) -> str:
        return "Transform(\n{},\n{}\n)".format(self.orient, self.pos)

    __repr__ = __str__

    @property
    def pos(self) -> Vector:
        """
        Access the position part of the matrix through a Vector object
        """
        return Vector(self._data[:3, 3], dtype=self.dtype, frozen=self._frozen)

    @pos.setter
    def pos(self, vector: Vector) -> None:
        if not isinstance(vector, Vector):
            raise ValueError()
        self._data[:3, 3] = vector.data

    @property
    def orient(self) -> Orientation:
        """
        Access the orientation part of the matrix through an Orientation object
        """
        return Orientation(self._data[:3, :3], dtype=self.dtype, frozen=self._frozen)

    @orient.setter
    def orient(self, orient: Orientation) -> None:
        if not isinstance(orient, Orientation):
            raise ValueError()
        self._data[:3, :3] = orient.data

    @property
    def data(self) -> np.ndarray:
        """
        Access the numpy array used by Transform
        """
        return self._data

    array = data
    matrix = data

    def inverse(self) -> Transform:
        """
        Return inverse of Transform
        """
        return Transform(matrix=np.linalg.inv(self._data))

    def invert(self) -> None:
        """
        In-place inverse the matrix
        """
        if self.frozen:
            raise ValueError("This Transform is frozen")
        self._data[:, :] = np.linalg.inv(self._data)

    def __eq__(self, other: object) -> bool:
        return self.similar(other)

    def __mul__(self, other: T) -> T:
        if isinstance(other, Vector):
            data = self.orient.data @ other.data + self.pos.data
            return Vector(data)  # type: ignore
        if isinstance(other, Transform):
            return Transform(matrix=self._data @ other.data)  # type: ignore
        if isinstance(other, np.ndarray):
            # This make it easy to support several format of point clouds but might be mathematically wrong
            if other.shape[1] == 3:
                return (self.orient.data @ other.T).T + self.pos.data
            if other.shape[0] == 3:
                return (self.orient.data @ other) + self.pos.data.reshape(3, 1)
            raise ValueError("Array shape must be 3, x or x, 3")
        return NotImplemented

    __matmul__ = __mul__

    @property
    def pose_vector(self) -> np.ndarray:
        return self.to_pose_vector()

    def to_pose_vector(self) -> np.ndarray:
        """
        Return a representation of transformation as 6 numbers array
        3 for position, and 3 for rotation vector
        """
        v = self.orient.to_rotation_vector()
        return np.array([self.pos.x, self.pos.y, self.pos.z, v.x, v.y, v.z])

    @classmethod
    def from_pose_vector(cls, x: NumberType, y: NumberType, z: NumberType, r1: NumberType, r2: NumberType, r3: NumberType) -> Transform:
        o = Orientation.from_rotation_vector(Vector(r1, r2, r3))
        return cls(o, [x, y, z])._validate()

    def to_ros(self):
        return self.orient.to_quaternion(), self.pos.data

    @classmethod
    def from_ros(cls, q, v):
        orient = Orientation.from_quaternion(*q)
        return cls(orient, Vector(v))._validate()

    def as_adjoint(self) -> np.ndarray:
        """
        Returns the 6x6 adjoint representation of the transform,
        that can be used to transform any 6-vector twist
        https://en.wikipedia.org/wiki/Adjoint_representation
        """
        return np.vstack(
            [
                np.hstack([self.orient.data, np.zeros((3, 3))]),
                np.hstack([np.dot(self.pos.as_so3(), self.orient.data), self.orient.data]),
            ]
        )

    @classmethod
    def from_corresponding_points(cls, fixed: np.ndarray, moving: np.ndarray) -> Transform:
        """
        Given a set of points and another set of points
        representing matching points of those in another coordinate
        system, compute a least squares transform between them using
        SVD

        """
        if fixed.shape != moving.shape:
            raise ValueError("input point clouds must be same length")

        if np.allclose(fixed, moving):
            return cls()

        centroid_f = np.mean(fixed, axis=0)
        centroid_m = np.mean(moving, axis=0)

        f_centered = fixed - centroid_f
        m_centered = moving - centroid_m

        B = f_centered.T @ m_centered

        # find rotation
        U, D, V = np.linalg.svd(B)
        R = V.T @ U.T

        # special reflection case
        if np.linalg.det(R) < 0:
            V[2, :] *= -1
            R = V.T @ U.T

        t = -R @ centroid_f + centroid_m

        return cls(Orientation(R), Vector(t))._validate()

    def copy(self) -> Transform:
        return Transform(matrix=self._data.copy())

    def dist(self, other: Transform) -> float:
        """
        Return distance equivalent between this matrix and a second one
        """
        return self.pos.dist(other.pos) + self.orient.ang_dist(other.orient)

    def similar(self, other: object, tol: NumberType = float_eps) -> bool:
        """
        Return True if distance to other transform is less than tol
        return False otherwise
        """
        if not isinstance(other, Transform):
            raise ValueError("Expecting a Transform object, received {} of type {}".format(other, type(other)))
        return bool(self.dist(other) <= tol)

    @staticmethod
    def mean(*transforms: Transform) -> Transform:
        return Transform(Orientation.mean(*(trf.orient for trf in transforms)), Vector.mean(*(trf.pos for trf in transforms)))

    def rotated_xb(self, angle: float) -> Transform:
        return Transform(Orientation.from_x_rotation(angle)) * self

    def rotated_yb(self, angle: float) -> Transform:
        return Transform(Orientation.from_y_rotation(angle)) * self

    def rotated_zb(self, angle: float) -> Transform:
        return Transform(Orientation.from_z_rotation(angle)) * self

    def rotated_xt(self, angle: float) -> Transform:
        return self * Transform(Orientation.from_x_rotation(angle))

    def rotated_yt(self, angle: float) -> Transform:
        return self * Transform(Orientation.from_y_rotation(angle))

    def rotated_zt(self, angle: float) -> Transform:
        return self * Transform(Orientation.from_z_rotation(angle))



import m3d
import numpy as np

CONVERTION_ERROR = 1000 * m3d.float_eps


def test_transform_mean():
    trfs = [
        m3d.Transform(
            m3d.Orientation(
                [
                    [1, 0, 0],
                    [0, 0, -1],
                    [0, 1, 0],
                ]
            ),
            m3d.Vector(1, 0, 0),
        ),
        m3d.Transform(
            m3d.Orientation(
                [
                    [0, 0, 1],
                    [0, 1, 0],
                    [-1, 0, 0],
                ]
            ),
            m3d.Vector(0, 1, 0),
        ),
        m3d.Transform(
            m3d.Orientation(
                [
                    [0, -1, 0],
                    [1, 0, 0],
                    [0, 0, 1],
                ]
            ),
            m3d.Vector(0, 0, 1),
        ),
    ]

    mean_trf = m3d.Transform.mean(*trfs)
    assert mean_trf.similar(
        m3d.Transform(
            m3d.Orientation(
                [
                    [2 / 3, -1 / 3, 2 / 3],
                    [2 / 3, 2 / 3, -1 / 3],
                    [-1 / 3, 2 / 3, 2 / 3],
                ]
            ),
            m3d.Vector(1 / 3, 1 / 3, 1 / 3),
        ),
        tol=CONVERTION_ERROR,
    )
    assert not mean_trf.similar(
        m3d.Transform.from_pose_vector(*np.mean([trf.pose_vector for trf in trfs], axis=0)),
        tol=CONVERTION_ERROR,
    )


def test_order():
    trfs = [
        m3d.Transform(
            m3d.Orientation(
                [
                    [1, 0, 0],
                    [0, 0, -1],
                    [0, 1, 0],
                ]
            ),
            m3d.Vector(1, 0, 0),
        ),
        m3d.Transform(
            m3d.Orientation(
                [
                    [0, 0, 1],
                    [0, 1, 0],
                    [-1, 0, 0],
                ]
            ),
            m3d.Vector(0, 1, 0),
        ),
        m3d.Transform(
            m3d.Orientation(
                [
                    [0, -1, 0],
                    [1, 0, 0],
                    [0, 0, 1],
                ]
            ),
            m3d.Vector(0, 0, 1),
        ),
    ]

    m3d.Transform.mean(*trfs).similar(m3d.Transform.mean(*trfs[::-1]))


def test_mean_simple_rotations():
    mean_ori = m3d.Orientation.mean(m3d.Orientation())
    assert mean_ori.similar(m3d.Orientation(), tol=CONVERTION_ERROR)

    mean_ori = m3d.Orientation.mean(*[m3d.Orientation()] * 5)
    assert mean_ori.similar(m3d.Orientation(), tol=CONVERTION_ERROR)

    mean_ori = m3d.Orientation.mean(
        *[
            m3d.Orientation.from_rotation_vector([0, 0, np.pi / 4]),
            m3d.Orientation.from_rotation_vector([0, 0, -np.pi / 4]),
        ]
    )
    assert mean_ori.similar(m3d.Orientation(), tol=CONVERTION_ERROR)

    mean_ori = m3d.Orientation.mean(
        *[
            m3d.Orientation.from_rotation_vector([0, 0, np.pi / 2]),
            m3d.Orientation.from_rotation_vector([0, 0, -np.pi / 2]),
        ]
    )
    assert mean_ori.similar(m3d.Orientation(), tol=CONVERTION_ERROR)


def test_rotation_direction():
    mean_ori = m3d.Orientation.mean(
        *[
            m3d.Orientation(),
            m3d.Orientation.from_rotation_vector([0, 0, 3 * np.pi / 2]),
        ]
    )
    assert mean_ori.similar(m3d.Orientation.from_rotation_vector([0, 0, 7 * np.pi / 4]), tol=CONVERTION_ERROR)

    mean_ori = m3d.Orientation.mean(
        *[
            m3d.Orientation(),
            m3d.Orientation.from_rotation_vector([0, 0, np.pi / 2]),
        ]
    )
    assert mean_ori.similar(m3d.Orientation.from_rotation_vector([0, 0, np.pi / 4]), tol=CONVERTION_ERROR)

    mean_ori = m3d.Orientation.mean(
        *[
            m3d.Orientation(),
            m3d.Orientation.from_rotation_vector([0, 0, np.pi]),
        ]
    )
    assert mean_ori.similar(m3d.Orientation.from_rotation_vector([0, 0, np.pi / 2]), tol=CONVERTION_ERROR) or mean_ori.similar(
        m3d.Orientation.from_rotation_vector([0, 0, 3 * np.pi / 2]), tol=CONVERTION_ERROR
    )

    mean_ori = m3d.Orientation.mean(
        *[
            m3d.Orientation(),
            m3d.Orientation.from_rotation_vector([0, 0, np.pi / 2]),
            m3d.Orientation.from_rotation_vector([0, 0, np.pi]),
        ]
    )
    assert mean_ori.similar(m3d.Orientation.from_rotation_vector([0, 0, np.pi / 2]), tol=CONVERTION_ERROR)

    mean_ori = m3d.Orientation.mean(
        *[
            m3d.Orientation(),
            m3d.Orientation.from_rotation_vector([0, 0, np.pi]),
            m3d.Orientation.from_rotation_vector([0, 0, 3 * np.pi / 2]),
        ]
    )
    assert mean_ori.similar(m3d.Orientation.from_rotation_vector([0, 0, 3 * np.pi / 2]), tol=CONVERTION_ERROR)


def test_equal_weights_vector():
    vectors = [
        m3d.Vector(0, 0, 0),
        m3d.Vector(2, 2, 2),
    ]

    assert m3d.Vector.mean(*vectors, weights=[1, 1]).similar(m3d.Vector.mean(*vectors))
    assert m3d.Vector.mean(*vectors, weights=[1, 1]).similar(m3d.Vector.mean(*vectors, weights=[2, 2]))


def test_zero_weight_vector():
    vectors = [
        m3d.Vector(0, 0, 0),
        m3d.Vector(2, 2, 2),
    ]

    assert m3d.Vector().similar(m3d.Vector.mean(*vectors, weights=[1, 0]))
    assert m3d.Vector().similar(m3d.Vector.mean(*vectors, weights=[0.5, 0]))


def test_weighted_vector():
    vectors = [
        m3d.Vector(0, 0, 0),
        m3d.Vector(3, 3, 3),
    ]
    weights = [2, 1]

    assert m3d.Vector.mean(*vectors, weights=weights).similar(m3d.Vector(1, 1, 1))



from __future__ import annotations
from typing import Union

import numpy as np

from m3d.common import float_eps
from m3d.transform import Transform
from m3d.quaternion import Quaternion


class DualQuaternion:
    def __init__(self, vector: np.ndarray) -> None:
        self._data = vector

    def __str__(self) -> str:
        return "DualQuaternion(\n{}, \n{})".format(self._data[:4], self._data[4:])

    __repr__ = __str__

    @property
    def data(self) -> np.ndarray:
        return self._data

    @property
    def q_rot(self) -> Quaternion:
        return Quaternion(self._data[:4])

    @property
    def q_trans(self) -> Quaternion:
        return Quaternion(self._data[4:])

    @classmethod
    def from_transformation(cls, trans: Transform) -> DualQuaternion:
        """
        Write the transformation as a dual quaternion. A dual quaternion can be
        written as q = r + epslion * d
        where the r quaternion is known as the real or rotational part and the
        d quaternion is known as the dual or displacement part. epsilon is the
        dual unit.
        We denote q_rot as the rotation part and q_trans as the displacement
        part.
        """

        q_rot = Quaternion(trans.orient.to_quaternion())
        q_trans = 0.5 * Quaternion([0, *trans.pos.data]) * q_rot
        return cls(np.array([*q_rot.data, *q_trans.data]))

    def __mul__(self, other: Union[float, int, DualQuaternion]) -> DualQuaternion:
        if isinstance(other, DualQuaternion):
            q_rot = self.q_rot * other.q_rot
            q_trans = self.q_rot * other.q_trans + self.q_trans * other.q_rot
            return DualQuaternion(np.array([*q_rot.data, *q_trans.data]))
        if isinstance(other, (float, int)):
            return DualQuaternion(other * self._data)
        raise ValueError()

    def __rmul__(self, other: Union[float, int]) -> DualQuaternion:
        if isinstance(other, (float, int)):
            return DualQuaternion(self._data * other)
        raise ValueError()

    @property
    def conjugate(self) -> DualQuaternion:
        return DualQuaternion(np.array([*self.q_rot.conjugate.data, *self.q_trans.conjugate.data]))

    @property
    def unity_condition(self) -> bool:
        """
        When working with rigid rotational and translational transformations
        the dual quaternion must fulfill the unit condition, i.e.
        it must be of unit length.
        """
        norm = np.linalg.norm((self * self.conjugate).data)
        return bool(abs(1 - norm) < float_eps)



from __future__ import annotations
from typing import Union

import numpy as np


class Quaternion:
    """
    Represent a quaternion
    Takes a list/array of length 4 as argument.
    default to [0, 0, 0 ,0]
    """

    def __init__(self, x=None):
        if x is None:
            x = [0, 0, 0, 0]
        if isinstance(x, (list, tuple)):
            if len(x) == 4:
                self._data = np.array(x)
            else:
                raise ValueError(f"A list of length 4 is expected, got {x}")
        elif isinstance(x, np.ndarray):
            if x.shape == (4,):
                self._data = x
            else:
                raise ValueError(f"A array of shape (4,) is expected got {x}")
        else:
            raise ValueError(f"A array of shape (4,) is expected got {x}")

    def __str__(self) -> str:
        return "Quaternion({})".format(self._data)

    __repr__ = __str__

    @property
    def data(self) -> np.ndarray:
        """
        Access the numpy array used by Quaternion
        """
        return self._data

    @property
    def scalar(self) -> float:
        return self._data[0]

    @property
    def vec(self) -> np.ndarray:
        return self._data[1:]

    def __add__(self, other: object) -> Quaternion:
        if isinstance(other, Quaternion):
            return Quaternion(self.data + other.data)
        raise ValueError(f"Cannot add quaternion and type {type(other)}")

    def __rmul__(self, other: object) -> Quaternion:
        if not isinstance(other, (float, int)):
            raise ValueError()
        return Quaternion(self._data * other)

    def __mul__(self, other: Union[int, float, Quaternion]) -> Quaternion:
        if isinstance(other, Quaternion):
            s1 = self._data[0]
            s2 = other._data[0]
            q1 = self._data[1:]
            q2 = other._data[1:]
            scalar_part = np.asarray([s1 * s2 - np.dot(q1, q2)])
            imag_part = np.asarray(s1 * q2 + s2 * q1 + np.cross(q1, q2))
            return Quaternion(np.hstack((scalar_part, imag_part)))
        if isinstance(other, (int, float)):
            return Quaternion(other * self._data)

        raise ValueError("Cannot multiply quaternion with type {}".format(type(other)))

    @property
    def conjugate(self) -> Quaternion:
        return Quaternion([self._data[0], -self._data[1], -self.data[2], -self.data[3]])

    def copy(self) -> Quaternion:
        return Quaternion(self._data.copy())



from __future__ import annotations
from typing import Optional, Sequence, Union

import numpy as np

from m3d.common import NumberType, float_eps, number_types


class Vector:
    """
    Represent a vector.
    Takes either x, y, z as argument or a list/array
    """

    def __init__(self, x=0.0, y=0.0, z=0.0, dtype=np.float32, frozen: bool = False):
        self._data: np.ndarray
        if isinstance(x, (list, tuple)):
            if len(x) == 3:
                self._data = np.array(x, dtype=dtype)
            else:
                raise ValueError(f"A list of length 3 is expected, got {x}")
        elif isinstance(x, np.ndarray):
            if x.shape == (3,):
                self._data = x
            else:
                raise ValueError(f"A array of shape (3,) is expected got {x}")
        else:
            self._data = np.array([float(x), float(y), float(z)], dtype=dtype)
        self._frozen = frozen
        self._data.flags.writeable = not frozen

    def __getitem__(self, idx: int) -> float:
        return self._data[idx]

    def __setitem__(self, idx: int, val: float) -> None:
        self._data[idx] = val

    @property
    def frozen(self) -> bool:
        return self._frozen

    @frozen.setter
    def frozen(self, val: bool) -> None:
        self._frozen = val
        self._data.flags.writeable = not val

    def copy(self) -> Vector:
        return Vector(self._data.copy())

    @property
    def x(self) -> float:
        return float(self._data[0])

    @x.setter
    def x(self, val: float) -> None:
        self._data[0] = val

    @property
    def y(self) -> float:
        return float(self._data[1])

    @y.setter
    def y(self, val: float) -> None:
        self._data[1] = val

    @property
    def z(self) -> float:
        return float(self._data[2])

    @z.setter
    def z(self, val: float) -> None:
        self._data[2] = val

    def __str__(self) -> str:
        return "Vector({}, {}, {})".format(self.x, self.y, self.z)

    def __sub__(self, other: Vector) -> Vector:
        return Vector(self.data - other.data)

    def __add__(self, other: Vector) -> Vector:
        return Vector(self.data + other.data)

    def __neg__(self) -> Vector:
        return Vector(-self._data)

    __repr__ = __str__

    @property
    def data(self) -> np.ndarray:
        return self._data

    array = data

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Vector):
            return NotImplemented
        return self.similar(other)

    def __mul__(self, other: NumberType) -> Vector:
        if not isinstance(other, number_types):
            raise ValueError()
        return Vector(self._data * other)

    __rmul__ = __mul__

    def __truediv__(self, other: NumberType) -> Vector:
        if not isinstance(other, number_types):
            raise ValueError()
        return Vector(self._data / other)

    def __itruediv__(self, other: NumberType) -> Vector:
        if not isinstance(other, number_types):
            raise ValueError()
        self._data /= other
        return self

    def __imul__(self, other: NumberType) -> Vector:
        if not isinstance(other, number_types):
            raise ValueError()
        self._data *= other
        return self

    @property
    def length(self) -> float:
        return float(np.linalg.norm(self.data))

    def dist(self, other: Vector) -> float:
        """
        return abolute distance to another vector
        """
        v = Vector(other.data - self.data)
        return v.length

    def similar(self, other: Vector, tol: NumberType = float_eps) -> bool:
        """
        Return True if distance to other Vector is less than tol
        return False otherwise
        """
        if not isinstance(other, Vector):
            raise ValueError("Expecting a Vector object, received {} of type {}".format(other, type(other)))
        return bool(self.dist(other) <= tol)

    def normalize(self) -> None:
        """
        Normalize in place vector
        """
        if self.length == 0:
            return
        self._data /= self.length

    def normalized(self) -> Vector:
        """
        Return a normalized copy of vector
        """
        if self.length == 0:
            return Vector(self.data)
        return Vector(self._data / self.length)

    def cross(self, other: Vector) -> Vector:
        if not isinstance(other, Vector):
            other = Vector(other)
        return Vector(np.cross(self.data, other.data))

    def dot(self, other: Vector) -> float:
        if not isinstance(other, Vector):
            other = Vector(other)
        return np.dot(self.data, other.data)

    __matmul__ = dot

    def project(self, other: Vector) -> Vector:
        if not isinstance(other, Vector):
            other = Vector(other)
        other = other.normalized()
        return self.dot(other) * other

    def angle(self, other: Vector, normal_vector: Optional[Vector] = None) -> float:
        """
        If provided, normal_vector is a vector defining the reference plane to be used to compute sign of angle.
        Otherwise, returned angle is between 0 and pi.
        """
        cos = self.dot(other) / (self.length * other.length)
        angle = np.arccos(np.clip(cos, -1, 1))
        if normal_vector is not None:
            angle = angle * np.sign(normal_vector.dot(self.cross(other)))
        return angle

    def as_so3(self) -> np.ndarray:
        """
        Returns the skew symetric (so3) representation of the vector
        https://en.wikipedia.org/wiki/Skew-symmetric_matrix
        """
        return np.array(
            [
                [0, -self.z, self.y],
                [self.z, 0, -self.x],
                [-self.y, self.x, 0],
            ]
        )

    @staticmethod
    def mean(*vectors: "Vector", weights: Optional[Union[Sequence[float], np.ndarray]] = None) -> "Vector":
        if weights is None:
            weights = np.ones(len(vectors))

        if len(vectors) != len(weights):
            raise ValueError("The number of weights needs to correspond to the number of vectors to average")
        if abs(sum(weights)) < float_eps:
            raise ValueError("Can not have all weights 0 or close to 0")

        return Vector(*np.average([vec.data for vec in vectors], axis=0, weights=weights))


# some units vectors
e0 = ex = Vector(1, 0, 0)
e1 = ey = Vector(0, 1, 0)
e2 = ez = Vector(0, 0, 1)



from __future__ import annotations
import math
from typing import List, Optional, Tuple, TypeVar

import numpy as np

from m3d.vector import Vector
from m3d.common import NumberType, float_eps


T = TypeVar("T")
UNIT_EIGENVECTOR_THRESHOLD = 1e-4


class Orientation:
    def __init__(self, data: Optional[np.ndarray] = None, dtype=np.float32, frozen: bool = False) -> None:
        if isinstance(data, np.ndarray):
            if data.shape == (3, 3):
                self._data = data
            else:
                raise ValueError(f"A numpy array of size (3, 3) is expected not {data.shape}")
        elif isinstance(data, list):
            self._data = np.array(data)
            if self._data.shape != (3, 3):
                raise ValueError(f"Creating an array from argument {data} did not lead to an array of shape (3, 3)")
        elif data is None:
            self._data = np.identity(3, dtype=dtype)
        else:
            raise ValueError(f"A numpy array of size (3, 3) is expected not {data}")
        self._frozen = frozen
        self._data.flags.writeable = not frozen

    @property
    def frozen(self) -> bool:
        return self._frozen

    @frozen.setter
    def frozen(self, val: bool) -> None:
        self._frozen = val
        self._data.flags.writeable = not val

    def is_valid(self) -> bool:
        """
        A real orthogonal matrix with det(R) = 1 provides a matrix representation of a proper
        rotation. Furthermore, a real orthogonal matrix with det (R) = −1 provides a matrix
        representation of an improper rotation.
        """
        if not np.allclose(self._data.T @ self._data, np.eye(3), float_eps, float_eps):
            return False
        if not (1 - np.linalg.det(self._data)) < float_eps:
            return False
        return True

    def _validate(self) -> Orientation:
        if not self.is_valid():
            raise ValueError("Given args gives an invalid Orientation")
        return self

    def rotate_xb(self, val: float) -> None:
        o = Orientation.from_x_rotation(val)
        self._data[:] = o.data @ self._data

    def rotate_yb(self, val: float) -> None:
        o = Orientation.from_y_rotation(val)
        self._data[:] = o.data @ self._data

    def rotate_zb(self, val: float) -> None:
        o = Orientation.from_z_rotation(val)
        self._data[:] = o.data @ self._data

    def rotate_xt(self, val: float) -> None:
        o = Orientation.from_x_rotation(val)
        self._data[:] = self._data @ o.data

    def rotate_yt(self, val: float) -> None:
        o = Orientation.from_y_rotation(val)
        self._data[:] = self._data @ o.data

    def rotate_zt(self, val: float) -> None:
        o = Orientation.from_z_rotation(val)
        self._data[:] = self._data @ o.data

    def rotated_xb(self, val: float) -> Orientation:
        return Orientation.from_x_rotation(val) * self

    def rotated_yb(self, val: float) -> Orientation:
        return Orientation.from_y_rotation(val) * self

    def rotated_zb(self, val: float) -> Orientation:
        return Orientation.from_z_rotation(val) * self

    def rotated_xt(self, val: float) -> Orientation:
        return self * Orientation.from_x_rotation(val)

    def rotated_yt(self, val: float) -> Orientation:
        return self * Orientation.from_y_rotation(val)

    def rotated_zt(self, val: float) -> Orientation:
        return self * Orientation.from_z_rotation(val)

    @classmethod
    def from_x_rotation(cls, angle: float) -> Orientation:
        return cls(
            np.array(
                [
                    [1, 0, 0],
                    [0, np.cos(angle), -np.sin(angle)],
                    [0, np.sin(angle), np.cos(angle)],
                ]
            )
        )

    @classmethod
    def from_y_rotation(cls, angle: float) -> Orientation:
        return cls(
            np.array(
                [
                    [np.cos(angle), 0, np.sin(angle)],
                    [0, 1, 0],
                    [-np.sin(angle), 0, np.cos(angle)],
                ]
            )
        )

    @classmethod
    def from_z_rotation(cls, angle: float) -> Orientation:
        return cls(
            np.array(
                [
                    [np.cos(angle), -np.sin(angle), 0],
                    [np.sin(angle), np.cos(angle), 0],
                    [0, 0, 1],
                ]
            )
        )

    def __str__(self) -> str:
        data = np.array2string(self.data, separator=", ")
        return "Orientation(\n{}\n)".format(data)

    __repr__ = __str__

    def inverse(self) -> Orientation:
        return Orientation(np.linalg.inv(self.data))

    def ang_dist(self, other: Orientation) -> float:
        r = self * other.inverse()
        trace_r = r.data[0, 0] + r.data[1, 1] + r.data[2, 2]
        if trace_r > 3:
            # might happen with approximations/rouding
            trace_r = 3
        return np.arccos((trace_r - 1) / 2)

    @property
    def data(self) -> np.ndarray:
        return self._data

    @property
    def array(self) -> np.ndarray:
        return self._data

    matrix = data

    def __mul__(self, other: T) -> T:
        if isinstance(other, Vector):
            return Vector(self._data @ other.data)  # type: ignore
        if isinstance(other, Orientation):
            return Orientation(self._data @ other.data)  # type: ignore
        if isinstance(other, np.ndarray):
            if other.shape[0] == 3:
                return self.data @ other  # type: ignore
            if other.shape[1] == 3:
                return (self.data @ other.T).T  # type: ignore
            raise ValueError("Array shape must be 3,x or x,3")
        return NotImplemented

    __matmul__ = __mul__

    def __eq__(self, other: object) -> bool:
        return self.similar(other)

    def to_quaternion(self) -> Tuple[float, float, float, float]:
        """
        Returns w, x, y, z
        adapted from
        https://github.com/matthew-brett/transforms3d/blob/master/transforms3d/quaternions.py
        """

        Qxx, Qyx, Qzx, Qxy, Qyy, Qzy, Qxz, Qyz, Qzz = self._data.flat
        # Fill only lower half of symmetric matrix
        K = (
            np.array(
                [
                    [Qxx - Qyy - Qzz, 0, 0, 0],
                    [Qyx + Qxy, Qyy - Qxx - Qzz, 0, 0],
                    [Qzx + Qxz, Qzy + Qyz, Qzz - Qxx - Qyy, 0],
                    [Qyz - Qzy, Qzx - Qxz, Qxy - Qyx, Qxx + Qyy + Qzz],
                ]
            )
            / 3.0
        )
        # Use Hermitian eigenvectors, values for speed
        vals, vecs = np.linalg.eigh(K)
        # Select largest eigenvector, reorder to w,x,y,z quaternion
        q = vecs[[3, 0, 1, 2], np.argmax(vals)]
        # Prefer quaternion with positive w
        # (q * -1 corresponds to same rotation as q)
        if q[0] < 0:
            q *= -1
        return q[0], q[1], q[2], q[3]

    @classmethod
    def from_quaternion(cls, w: NumberType, x: NumberType, y: NumberType, z: NumberType) -> Orientation:
        # adapted from
        # https://github.com/matthew-brett/transforms3d/blob/master/transforms3d/quaternions.py
        Nq = w * w + x * x + y * y + z * z
        if Nq < float_eps:
            return Orientation()
        s = 2.0 / Nq
        X = x * s
        Y = y * s
        Z = z * s
        wX = w * X
        wY = w * Y
        wZ = w * Z
        xX = x * X
        xY = x * Y
        xZ = x * Z
        yY = y * Y
        yZ = y * Z
        zZ = z * Z
        return cls(
            np.array(
                [
                    [1.0 - (yY + zZ), xY - wZ, xZ + wY],
                    [xY + wZ, 1.0 - (xX + zZ), yZ - wX],
                    [xZ - wY, yZ + wX, 1.0 - (xX + yY)],
                ]
            )
        )._validate()

    @classmethod
    def from_axis_angle(cls, axis: Vector, angle: float, is_normalized: bool = False) -> Orientation:
        # adapted from
        # https://github.com/matthew-brett/transforms3d/blob/master/transforms3d/quaternions.py

        axis = Vector(axis) if not isinstance(axis, Vector) else axis

        if not is_normalized:
            axis.normalize()
        x, y, z = axis.x, axis.y, axis.z
        c = math.cos(angle)
        s = math.sin(angle)
        C = 1 - c
        xs = x * s
        ys = y * s
        zs = z * s
        xC = x * C
        yC = y * C
        zC = z * C
        xyC = x * yC
        yzC = y * zC
        zxC = z * xC
        return cls(
            np.array(
                [
                    [x * xC + c, xyC - zs, zxC + ys],
                    [xyC + zs, y * yC + c, yzC - xs],
                    [zxC - ys, yzC + xs, z * zC + c],
                ]
            )
        )._validate()

    @classmethod
    def from_xy(cls, x_vec: Vector, y_vec: Vector) -> Orientation:
        """
        Generate a new Orientation from two vectors using x as reference
        """
        if not isinstance(x_vec, Vector):
            x_vec = Vector(x_vec)
        if not isinstance(y_vec, Vector):
            y_vec = Vector(y_vec)
        x_vec.normalize()
        y_vec.normalize()
        orient = cls()
        orient.data[:, 0] = x_vec.data
        orient.data[:, 2] = x_vec.cross(y_vec).normalized().data
        orient.data[:, 1] = Vector(np.cross(orient.data[:, 2], x_vec.data)).normalized().data
        return orient._validate()

    @classmethod
    def from_yz(cls, y_vec: Vector, z_vec: Vector) -> Orientation:
        """
        Generate a new Orientation from two vectors using y as reference
        """
        if not isinstance(y_vec, Vector):
            y_vec = Vector(y_vec)
        if not isinstance(z_vec, Vector):
            z_vec = Vector(z_vec)
        y_vec.normalize()
        z_vec.normalize()
        orient = cls()
        orient.data[:, 1] = y_vec.data
        orient.data[:, 0] = y_vec.cross(z_vec).normalized().data
        orient.data[:, 2] = Vector(np.cross(orient.data[:, 0], y_vec.data)).normalized().data
        return orient._validate()

    @classmethod
    def from_xz(cls, x_vec: Vector, z_vec: Vector, ref: str = "x") -> Orientation:
        """
        Generate a new Orientation from two vectors using x as reference
        """
        if not isinstance(x_vec, Vector):
            x_vec = Vector(x_vec)
        if not isinstance(z_vec, Vector):
            z_vec = Vector(z_vec)
        x_vec.normalize()
        z_vec.normalize()
        orient = cls()
        orient.data[:, 1] = z_vec.cross(x_vec).normalized().data

        if ref == "x":
            orient.data[:, 0] = x_vec.data
            orient.data[:, 2] = Vector(np.cross(x_vec.data, orient.data[:, 1])).normalized().data
        elif ref == "z":
            orient.data[:, 2] = z_vec.data
            orient.data[:, 0] = Vector(np.cross(orient.data[:, 1], z_vec.data)).normalized().data
        else:
            raise ValueError("Value of ref can only be x or z")

        return orient._validate()

    def to_axis_angle(self, unit_thresh: float = UNIT_EIGENVECTOR_THRESHOLD) -> Tuple[Vector, float]:
        # adapted from
        # https://github.com/matthew-brett/transforms3d/blob/master/transforms3d/quaternions.py
        M = np.asarray(self._data, dtype=np.float32)
        # direction: unit eigenvector of R33 corresponding to eigenvalue of 1
        L, W = np.linalg.eig(M.T)
        i = np.where(np.abs(L - 1.0) < unit_thresh)[0]
        if i.size == 0:
            raise ValueError("no unit eigenvector corresponding to eigenvalue 1")
        direction = np.real(W[:, i[-1]]).squeeze()
        # rotation angle depending on direction
        cosa = (np.trace(M) - 1.0) / 2.0
        if abs(direction[2]) > 1e-8:
            sina = (M[1, 0] + (cosa - 1.0) * direction[0] * direction[1]) / direction[2]
        elif abs(direction[1]) > 1e-8:
            sina = (M[0, 2] + (cosa - 1.0) * direction[0] * direction[2]) / direction[1]
        else:
            sina = (M[2, 1] + (cosa - 1.0) * direction[1] * direction[2]) / direction[0]
        angle = math.atan2(sina, cosa)
        return Vector(direction), angle

    def to_rotation_vector(self, unit_thresh: float = UNIT_EIGENVECTOR_THRESHOLD) -> Vector:
        v, a = self.to_axis_angle(unit_thresh)
        return v * a

    @classmethod
    def from_rotation_vector(cls, v: Vector) -> Orientation:
        if isinstance(v, (np.ndarray, list, tuple)):
            v = Vector(*v)
        if not isinstance(v, Vector):
            raise ValueError("Method take a Vector as argument")
        if v.length == 0:
            return cls(np.identity(3, dtype=np.float32))
        u = v.normalized()
        idx = (u.data != 0).argmax()
        return cls.from_axis_angle(u, v[idx] / u[idx], True)._validate()

    def copy(self) -> Orientation:
        return Orientation(self.data.copy())

    def similar(self, other: object, tol: float = float_eps) -> bool:
        """
        Return True if angular distance to other Orientation is less than tol
        return False otherwise
        """
        if not isinstance(other, Orientation):
            raise ValueError("Expecting an Orientation object, received {} of type {}".format(other, type(other)))
        return bool(self.ang_dist(other) <= tol)

    @property
    def vec_x(self) -> Vector:
        return Vector(self._data[:, 0])

    @property
    def vec_y(self) -> Vector:
        return Vector(self._data[:, 1])

    @property
    def vec_z(self) -> Vector:
        return Vector(self._data[:, 2])

    @staticmethod
    def mean(*orientations: Orientation) -> Orientation:
        """
        Averaging quaternions
        https://ntrs.nasa.gov/api/citations/20070017872/downloads/20070017872.pdf
        """
        try:
            from scipy.optimize import minimize, NonlinearConstraint
        except ImportError:
            raise Exception("scipy must be installed to use this method")
        arit_mean = np.mean([ori.to_rotation_vector().data for ori in orientations], axis=0).tolist()
        arit_mean_ori = Orientation.from_rotation_vector(arit_mean)
        x0 = np.array(arit_mean_ori.to_quaternion())

        quats: List[Tuple[float, float, float, float]] = [ori.to_quaternion() for ori in orientations]
        ans = minimize(
            func_error_matrix,
            x0,
            args=(quats),
            method="SLSQP",
            constraints=NonlinearConstraint(np.linalg.norm, 1 - 1e-20, 1 + 1e-20),
        )

        return Orientation.from_quaternion(*ans.x)


def func_error_matrix(x0: np.ndarray, quats: List[Tuple[float, float, float, float]]) -> float:
    total_error: float = 0.0
    for q in quats:
        vec = np.array(q[:3])
        scal = q[3]
        I3 = np.eye(3)
        qX = np.array(
            [
                [0, -q[2], q[1]],
                [q[2], 0, -q[0]],
                [-q[1], q[0], 0],
            ]
        )

        err_mtx = np.vstack((scal * I3 + qX, -vec.T))

        total_error += float(np.linalg.norm(np.matmul(err_mtx.T, x0)) ** 2)

    return total_error



import numpy as np

import pytest
import m3d

# when exporting and reimporting a transform to another expression we must accept a higher error then eps
CONVERTION_ERROR = 1000 * m3d.float_eps


def _are_equals(m1, m2, eps=m3d.float_eps):
    """
    to test equality of two objects with a tolerance
    """
    if isinstance(m1, (float, np.float32)) and isinstance(m2, (float, np.float32)):
        return abs(m1 - m2) < eps

    m1 = _to_np(m1)
    m2 = _to_np(m2)

    return (abs(m1 - m2) <= eps).all()


def _to_np(obj):
    if isinstance(obj, np.ndarray):
        return obj
    elif isinstance(obj, (m3d.Vector, m3d.Orientation, m3d.Transform)):
        return obj.data
    else:
        raise ValueError("Could not convert obj to nupy array", obj, type(obj))


def test_init():
    t = m3d.Transform()
    assert t.pos.x == 0
    assert t.pos.y == 0
    t.pos.x = 2
    assert t.pos.x == 2

    i = t.inverse()
    assert m3d.Vector(-2, 0, 0) == i.pos
    assert m3d.Orientation() == i.orient


def test_transform_init():
    v = m3d.Vector(1, 2, 3)
    o = m3d.Orientation()
    o.rotate_zb(1)
    t = m3d.Transform(o, v)
    assert t.pos == v
    assert t.orient == o
    t2 = m3d.Transform(o.data, v.data)
    assert t2.pos == v
    assert t2.orient == o
    assert t == t2
    with pytest.raises(ValueError):
        m3d.Transform(np.array([1, 2, 3]), np.array([1, 2, 3]))
    with pytest.raises(ValueError):
        m3d.Transform(o.data, np.array([1, 2]))


def test_transform_init_ref():
    v = m3d.Vector(1, 2, 3)
    o = m3d.Orientation()
    o.rotate_zb(1)
    t = m3d.Transform(o, v)
    assert t.orient == o
    o.rotate_xb(1)
    assert t.orient != o


def test_rotation():
    t = m3d.Transform()
    t.pos.x = 1
    t.orient.rotate_yb(1)
    res = m3d.Transform(
        m3d.Orientation(
            [
                [0.54030228, 0, 0.84147096],
                [0.0, 1, 0.0],
                [-0.84147096, 0.0, 0.54030228],
            ]
        ),
        m3d.Vector(1, 0, 0),
    )
    assert _are_equals(res.data, t.data)


def test_multiplication_orient():
    o = m3d.Orientation()
    o.rotate_zb(np.pi / 2)
    v = m3d.Vector(1, 0, 0)
    r = o * v
    assert r == m3d.Vector(0, 1, 0)
    o.rotate_zb(-np.pi)
    v = m3d.Vector(2, 0, 0)
    r = o * v
    assert r == m3d.Vector(0, -2, 0)


def test_transform() -> None:
    t = m3d.Transform()
    t.orient.rotate_zb(np.pi / 2)
    t.pos.y = 2
    v = m3d.Vector(1, 0, 0)
    r = t * v
    assert r == m3d.Vector(0, 3, 0)


def test_pose_vector():
    t = m3d.Transform()
    t.pos.x = 1
    t.pos.z = 2
    t.orient.rotate_yb(1)
    v = t.to_pose_vector()
    t2 = t.from_pose_vector(*v)
    assert t == t2


def test_rotation_vector():
    o = m3d.Orientation()
    o.rotate_yb(1)
    o.rotate_zb(1)
    v = o.to_rotation_vector()
    o2 = m3d.Orientation.from_rotation_vector(v)
    assert o.similar(o2, CONVERTION_ERROR)


def test_rotation_vector_2():
    o = m3d.Orientation()
    o.rotate_yb(1.1)
    v = o.to_rotation_vector()
    assert _are_equals(v[1], 1.1)
    o = m3d.Orientation()
    o.rotate_zb(1.1)
    v = o.to_rotation_vector()
    assert _are_equals(v[2], 1.1)
    o = m3d.Orientation()
    o.rotate_xb(1.1)
    v = o.to_rotation_vector()
    assert _are_equals(v[0], 1.1)


def test_vec_eq():
    a = m3d.Vector(-1.0000001192092896, -1.9999998807907104, -3.0)
    b = m3d.Vector(-1.0, -2.0, -3.0)
    assert a == b


def test_mult_trans():
    t1 = m3d.Transform()
    t1.orient.rotate_xb(np.pi / 2)
    t1.pos.x = 1

    t2 = m3d.Transform()
    t2.orient.rotate_xb(np.pi / 2)
    t2.pos.x = 2

    v = m3d.Vector(0, 0, 3)

    tr = m3d.Transform()
    tr.orient.rotate_xb(np.pi)
    tr.pos.x = 3

    assert t1 * t2 * v == tr * v
    assert t1 @ t2 @ v == tr @ v


def test_equal():
    t1 = m3d.Transform()
    t1.orient.rotate_xb(np.pi / 2)
    t1.pos.x = 1

    t2 = m3d.Transform()
    t2.orient.rotate_xb(np.pi / 2)
    t2.pos.x = 2

    tr = m3d.Transform()
    tr.orient.rotate_xb(np.pi)
    tr.pos.x = 3

    assert t1 != t2
    assert t1 != tr
    assert t1 * t2 == tr
    assert t2 * t1 == tr


def test_inverse_orient():
    o = m3d.Orientation()
    o.rotate_xb(3)
    o.rotate_yb(1)
    o.rotate_xb(1)
    v = m3d.Vector(-1, -2, -3)
    assert o * v != v
    assert o.inverse() * o * v == v
    assert o * o.inverse() * v == v
    assert o * o.inverse() * o == o


def test_inverse_trans():
    t = m3d.Transform()
    t.pos.x = 1
    t.orient.rotate_zb(np.pi / 2)
    t.orient.rotate_yb(np.pi / 2)
    v = m3d.Vector(2, 0, 0)
    assert t * v == m3d.Vector(1, 2, 0)


def test_inverse_invert():
    t = m3d.Transform()
    t.orient.rotate_xb(np.pi / 3)
    t.pos.x = 1
    t1 = t.copy()
    t2 = t1.inverse()
    n = t1.invert()
    assert n is None
    assert _are_equals(t1, t2)
    assert _are_equals(t * t1, m3d.Transform())


def test_inverse():
    t1 = m3d.Transform()
    t1.orient.rotate_xb(np.pi / 3)
    t1.pos.x = 1

    t2 = m3d.Transform()
    t2.orient.rotate_xb(-13 * np.pi / 6)
    t2.pos.x = 2.3

    v = m3d.Vector(0.1, -4.5, 3.0)

    tr = m3d.Transform()
    tr.orient.rotate_xb(np.pi)
    tr.pos.x = 3

    assert (t1 * t1.inverse()) == m3d.Transform(matrix=np.identity(4))
    assert (t1 * t1.inverse()) == m3d.Transform()
    assert (t2 * t2.inverse()) == m3d.Transform(matrix=np.identity(4))
    assert (t2 * t2.inverse()) == m3d.Transform()
    assert (t1 * t2 * t1.inverse() * t2.inverse()).similar(m3d.Transform(matrix=np.identity(4)), CONVERTION_ERROR)
    assert t1.inverse() * (t1 * v) == v


def test_inverse_2():
    t = m3d.Transform()
    t.pos.x = 1
    t.pos.z = 3
    t.orient.rotate_yb(1)
    t.orient.rotate_zb(1)
    v = m3d.Vector(2, 3, 4)
    assert v != t * v
    assert v == t.inverse() * t * v
    assert v == t * t.inverse() * v
    assert v != t @ v
    assert v == t.inverse() @ t @ v
    assert v == t @ t.inverse() @ v


def test_rotation_seq():
    t = m3d.Transform()
    t.pos.x = 1
    t.pos.z = 3
    t.orient.rotate_xb(1)
    res = t.copy()
    t.orient.rotate_yb(2)
    t.orient.rotate_zb(3)
    t.orient.rotate_zb(-3)
    t.orient.rotate_yb(-2)
    assert t == res


def test_rotation_seq_2():
    t = m3d.Transform()
    t.pos.x = 1
    t.pos.z = 3
    t.orient.rotate_xb(1)
    t.orient.rotate_yb(2)
    t.orient.rotate_zb(3)

    b = m3d.Transform()
    b.orient.rotate_zb(-3)
    b.orient.rotate_yb(-2)
    b.orient.rotate_xb(-1)
    b.pos = b.orient * m3d.Vector(1, 0, 3) * -1

    assert _are_equals(t.inverse().data, b.data)


def test_rotation_t():
    t = m3d.Transform()
    t.pos.x = 1
    t.orient.rotate_zt(np.pi / 2)
    t.orient.rotate_yt(np.pi / 2)
    v = m3d.Vector(2, 0, 0)
    assert t * v == m3d.Vector(1, 0, -2)


def test_rotation_t_2():
    t = m3d.Transform()
    t.orient.rotate_yt(-np.pi / 2)
    t.orient.rotate_xt(np.pi / 3)
    v = m3d.Vector(2, 0, 0)
    assert t * v == m3d.Vector(0, 0, 2)


def test_construct():
    o = m3d.Orientation()
    o.rotate_zb(1)
    v = m3d.Vector()
    v[0] = 1
    v[2] = 2
    t = m3d.Transform(o, v)
    assert t.pos.x == 1
    assert t.pos.z == 2
    t.pos = m3d.Vector()
    t.orient.rotate_zb(-1)
    assert t == m3d.Transform()
    t.orient = o
    assert t != m3d.Transform()


def test_orient():
    o = m3d.Orientation()
    o.rotate_zb(2)
    o2 = m3d.Orientation()
    o2.rotate_zb(2 * np.pi)
    assert o * o2 == o


def test_quaternion():
    o = m3d.Orientation()
    o.rotate_xb(np.pi / 3)
    o.rotate_zb(np.pi / 3)
    q = o.to_quaternion()
    o2 = m3d.Orientation.from_quaternion(*q)
    assert o.similar(o2, CONVERTION_ERROR)
    o = m3d.Orientation()
    o2 = m3d.Orientation.from_quaternion(0.0, 0.0, 0.0, 0.0)
    assert o.similar(o2, CONVERTION_ERROR)


def test_axis_angle():
    o = m3d.Orientation()
    o.rotate_xb(np.pi / 3)
    o.rotate_zb(np.pi / 3)
    v, a = o.to_axis_angle()
    o2 = m3d.Orientation.from_axis_angle(v, a)
    assert o.similar(o2, CONVERTION_ERROR)


def test_from_axis_angle():
    axis_tuple = (1, 2, 3)
    axis_normalized = m3d.Vector(axis_tuple).normalized()
    angle = 10

    o1 = m3d.Orientation.from_axis_angle(axis_tuple, angle)
    o2 = m3d.Orientation.from_axis_angle(axis_normalized, angle)
    assert o1.similar(o2, CONVERTION_ERROR)


def test_pc():
    pc = np.array([[1, 2, 3], [1, 2, 3], [2, 3, 4], [3, 4, 5], [3, 4, 5]])
    pc = pc.T
    t = m3d.Transform()
    t.pos.x = 1.2
    t.pos.y = 1
    t.orient.rotate_yb(1)
    t.orient.rotate_zb(1)

    a = t * pc
    assert a.shape == pc.shape

    c = t * pc.T
    assert c.shape == pc.T.shape


def test_copy():
    t = m3d.Transform()
    t.pos.x = 1
    new = t.copy()
    t.orient.rotate_zb(1)
    new.pos.x = 5
    assert t.pos.x != new.pos.x
    assert t.orient.data[0, 0] != new.orient.data[0, 0]
    v = t.pos.copy()
    assert v == t.pos
    v[1] = 3.33
    assert v.y != t.pos.y
    assert v != t.pos


def test_vect_mul():
    v1 = m3d.Vector(10, 10, 10)
    const_val = 2
    const_val_2 = -5
    v_res = v1 * const_val
    v_res_2 = v1 * const_val_2
    assert v_res == m3d.Vector(20, 20, 20)
    assert v_res_2 == m3d.Vector(-50, -50, -50)


def test_vect_div():
    v1 = m3d.Vector(10, 10, 10)
    const_val = 2
    const_val_2 = -5
    v_res = v1 / const_val
    v_res_2 = v1 / const_val_2
    assert v_res == m3d.Vector(5, 5, 5)
    assert v_res_2 == m3d.Vector(-2, -2, -2)


def test_substraction():
    v1 = m3d.Vector(1, 2, 3)
    v2 = m3d.Vector(2, -3, 4)
    v_res = v2 - v1
    assert v_res == m3d.Vector(1, -5, 1)


def test_addition():
    v1 = m3d.Vector(1, 2, 3)
    v2 = m3d.Vector(2, -3, 4)
    v_res = v2 + v1
    assert v_res == m3d.Vector(3, -1, 7)


def test_dist():
    v1 = m3d.Vector(1, 1, 1)
    v2 = m3d.Vector(2, 2, 2)
    v_res = v2.dist(v1)
    assert v_res == m3d.Vector(1, 1, 1).length


def test_ang_dist():
    o1 = m3d.Orientation()
    o1.rotate_yb(1)
    o1.rotate_xb(1)
    o1.rotate_zb(1)

    o2 = m3d.Orientation()
    o2.rotate_yb(2)
    o2.rotate_xb(2)
    o2.rotate_zb(2)

    o_dist_m3d = o1.ang_dist(o2)
    o_inv_dist_m3d = o2.ang_dist(o1)

    assert abs(o_dist_m3d - o_inv_dist_m3d) <= m3d.float_eps


def test_trans_dist():
    a = m3d.Transform()
    b = m3d.Transform()
    b.pos.x = 3
    b.orient.rotate_zb(1)
    b.orient.rotate_yb(1)
    assert a.dist(b) > 0.1
    assert b.dist(b) == 0


def test_eq():
    t = m3d.Transform()
    t.orient.rotate_yb(1)
    t.orient.rotate_zb(1)
    t.pos.x = 1
    v = m3d.Vector()
    o = m3d.Orientation()
    with pytest.raises(ValueError):
        assert t != v
    with pytest.raises(ValueError):
        assert v != t
    with pytest.raises(ValueError):
        assert o != v
    t2 = t.copy()
    assert t == t2
    assert t.pos == t2.pos
    assert t.orient == t2.orient
    t2.pos.y = 2
    assert t != t2
    assert t.pos != t2.pos
    assert t.orient == t2.orient
    t3 = t.copy()
    t3.orient.rotate_xb(1)
    assert t2.pos != t3.pos
    assert t2.orient != t3.orient
    assert t != t3


def test_norm():
    v = m3d.Vector(1, 2, 3)
    v.normalize()
    assert abs(v.length - 1) <= m3d.float_eps


def test_from_xy():
    x = m3d.Vector(1, 0, 0)
    y = m3d.Vector(0.01, 2.1, 0)
    orient = m3d.Orientation.from_xy(x, y)
    assert _are_equals(orient.data, np.identity(3), eps=0.1)


def test_from_yz():
    y = m3d.Vector(0, 1, 0)
    z = m3d.Vector(0, 0.01, 0.1)
    orient = m3d.Orientation.from_yz(y, z)
    assert _are_equals(orient.data, np.identity(3), eps=0.1)


def test_from_xz():
    x = m3d.Vector(1, 0, 0)
    z = m3d.Vector(0, 0, 1)
    orient = m3d.Orientation.from_xz(x, z)
    assert _are_equals(orient.data, np.identity(3))

    x = m3d.Vector(2, 0, 0.1)
    z = m3d.Vector(0.1, -0.1, 3)
    o = m3d.Orientation()
    o.rotate_yb(1)
    o.rotate_zb(1)
    x = o @ x
    z = o @ z
    orient = m3d.Orientation.from_xz(x, z)
    assert _are_equals(o, orient, eps=0.1)


def test_ros():
    t = m3d.Transform()
    t.orient.rotate_yb(1)
    t.orient.rotate_zb(1)
    t.pos.x = 1
    q, v = t.to_ros()
    t1 = t.from_ros(q, v)
    assert t == t1


def test_null_rotation_vector():
    o = m3d.Orientation.from_rotation_vector((0, 0, 0))
    assert np.array_equal(o.data, np.identity(3))


def test_update_trans_xyz():
    t = m3d.Transform()
    t.orient.rotate_yb(1)
    t.orient.rotate_zb(1)
    t.pos.x = 1

    t.pos.x += 1.2
    assert _are_equals(t.pos.x, 2.2)
    assert _are_equals(t.data[0, 3], 2.2)
    t.pos.y += 1.2
    assert _are_equals(t.pos.y, 1.2)
    assert _are_equals(t.data[1, 3], 1.2)


def test_similar():
    t = m3d.Transform()
    t.pos.x = 1
    t.pos.y = 2
    t.pos.z = 3
    t.orient.rotate_yb(1)

    t2 = t.copy()
    t2.orient.rotate_zb(1)

    t3 = t.copy()
    t3.orient.rotate_zb(1 - 4 * np.pi)

    t4 = t3.copy()
    t4.pos.x += m3d.float_eps

    t5 = t4.copy()
    t5.pos.x += 0.1

    assert not t.similar(t2)
    assert t2.similar(t3)
    assert t3.similar(t4)
    assert t2.similar(t4)
    assert t4.similar(t2)
    assert t4.orient.similar(t2.orient)
    assert not t.similar(t4)
    assert not t.orient.similar(t4.orient)
    assert not t4.similar(t)
    assert not t4.similar(t5)
    assert not t.similar(t5)
    assert t4.pos.similar(t5.pos, 0.2)
    assert t4.orient.similar(t5.orient, 0.2)
    assert t4.similar(t5, 0.2)


def test_orient_except():
    with pytest.raises(ValueError):
        o = m3d.Orientation(np.identity(4))
    with pytest.raises(ValueError):
        o = m3d.Orientation("whatever")
    o = m3d.Orientation()
    with pytest.raises(ValueError):
        o * np.identity(4)


def test_vector_except():
    with pytest.raises(ValueError):
        m3d.Vector(np.identity(4))
    with pytest.raises(ValueError):
        m3d.Vector("whatever")
    with pytest.raises(ValueError):
        m3d.Vector([1, 2, 3, 4])
    m3d.Vector([1, 2, 3])


def test_trans_init_except():
    with pytest.raises(ValueError):
        m3d.Vector(np.identity(5))


def test_unit_vectors():
    assert m3d.vector.e0 == m3d.Vector(1, 0, 0)
    assert m3d.vector.e1 == m3d.Vector(0, 1, 0)
    assert m3d.vector.e2 == m3d.Vector(0, 0, 1)
    t = m3d.Transform()
    assert m3d.vector.ex == t.orient.vec_x
    assert m3d.vector.ey == t.orient.vec_y
    assert m3d.vector.ez == t.orient.vec_z
    t.orient.rotate_zb(1)
    assert m3d.vector.ex != t.orient.vec_x
    assert m3d.vector.ey != t.orient.vec_y
    assert m3d.vector.ez == t.orient.vec_z


def test_vector_dot():
    d1 = np.array([1, 2, -3])
    d2 = np.array([2, 2, 2])
    v1 = m3d.Vector(d1)
    v2 = m3d.Vector(d2)
    assert np.dot(d1, d2) == v1.dot(v2)
    assert v1.dot(v2) == v1 @ v2
    assert v1 @ v1 == v1.length**2


def test_vector_project():
    v1 = m3d.Vector(1, 1, 1)
    v2 = m3d.Vector(2, 2, 2)

    vx = m3d.Vector(1, 0, 0)
    vy = m3d.Vector(0, 1, 0)
    vz = m3d.Vector(0, 0, 1)

    assert v1.project(vx) == vx
    assert v1.project(vy) == vy
    assert v1.project(vz) == vz

    assert v1.project(v2) == v1
    assert v2.project(v1) == v2


def test_vector_angle():
    v1 = m3d.Vector(1, 1, 1)
    v2 = m3d.Vector(2, 2, 2)

    vx = m3d.Vector(1, 0, 0)
    vy = m3d.Vector(0, 1, 0)
    vz = m3d.Vector(0, 0, 1)

    assert vx.angle(vx) == 0
    assert vx.angle(vy) == np.pi / 2
    assert vx.angle(vz) == np.pi / 2

    assert v1.angle(v2) == 0
    assert v2.angle(v1) == 0


def test_vector_angle_perp():
    v1 = m3d.Vector(1, 0, 0)
    v2 = m3d.Vector(1, 1, 0)
    v3 = m3d.Vector(1, -1, 0)

    vx = m3d.Vector(1, 0, 0)
    vy = m3d.Vector(0, 1, 0)
    vz = m3d.Vector(0, 0, 1)

    assert vx.angle(vx, vz) == 0
    assert vx.angle(vy, vz) == np.pi / 2
    assert vx.angle(vz, vy) == -np.pi / 2

    # Flip nominal direction for perpendicular vector
    assert vx.angle(vx, -vz) == 0
    assert vx.angle(vy, -vz) == -np.pi / 2
    assert vx.angle(vz, -vy) == np.pi / 2

    assert v1.angle(v2, vz) == -v1.angle(v2, -vz)
    assert v1.angle(v2, vz) == v1.angle(v3, -vz)


def test_quaternion_class():
    o = m3d.Orientation()
    o.rotate_xb(np.pi / 3)
    o.rotate_zb(np.pi / 3)
    q = o.to_quaternion()
    q_m3d = m3d.Quaternion(q)

    np.testing.assert_array_equal(np.array(q), q_m3d.data)


def test_dual_quaternion_unit_condition():
    for _ in range(20):
        t = m3d.Transform.from_pose_vector(*np.random.rand(6))
        dq = m3d.DualQuaternion.from_transformation(t)
        assert dq.unity_condition


def test_frozen_vector():
    v = m3d.Vector(1, 2, 3, frozen=True)
    with pytest.raises(ValueError):
        v.x = 1
    with pytest.raises(ValueError):
        v /= 2
    v.frozen = False
    v.x = 9
    v.frozen = True
    with pytest.raises(ValueError):
        v.x = 1
    assert v.x == 9


def test_frozen_transform():
    t = m3d.Transform(frozen=True)
    with pytest.raises(ValueError):
        t.orient.rotate_yb(1)
    with pytest.raises(ValueError):
        t.pos.x = 1
    t.frozen = False
    t.pos.x = 1
    t.frozen = True
    with pytest.raises(ValueError):
        t.pos.x = 2
    assert t.pos.x == 1
    t.frozen = False


def test_frozen_orient():
    o = m3d.Orientation(frozen=True)
    with pytest.raises(ValueError):
        o.rotate_yb(1)
    o.frozen = False
    o.rotate_yb(1)
    o.frozen = True
    with pytest.raises(ValueError):
        o.rotate_yb(1)


def test_adjoint_transformation():
    """
    The correct values are and can be deduced intuitively
    """
    ft = np.array([1, 2, 3, 0.1, 0.2, 0.3])
    trf = m3d.Transform(
        matrix=np.array(
            [
                [1, 0, 0, 0],
                [0, -1, 0, 0],
                [0, 0, -1, 0],
                [0, 0, 0, 1],
            ]
        )
    )

    np.testing.assert_almost_equal(
        trf.as_adjoint() @ ft,
        np.array([ft[0], -ft[1], -ft[2], ft[3], -ft[4], -ft[5]]),
    )

    x = 0.1
    y = 0.2
    trf = m3d.Transform(
        matrix=np.array(
            [
                [1, 0, 0, x],
                [0, 1, 0, y],
                [0, 0, 1, 0],
                [0, 0, 0, 1],
            ]
        )
    )

    np.testing.assert_almost_equal(
        trf.as_adjoint() @ ft,
        np.array([ft[0], ft[1], ft[2], ft[3] + y * ft[2], ft[4] - x * ft[2], ft[5] + x * ft[1] - y * ft[0]]),
    )


def test_valid_nan():
    t = m3d.Transform()
    assert t.is_valid()
    t._data[1, 0] = np.nan
    assert not t.is_valid()


def test_valid_0():
    t = m3d.Transform()
    assert t.is_valid()
    t._data[3, 1] = 0.01
    assert not t.is_valid()


def test_valid_1():
    t = m3d.Transform()
    assert t.is_valid()
    t._data[3, 3] = 0.99
    assert not t.is_valid()


def test_valid_and_invalid_rotation_matrices():
    reflection = m3d.Orientation(
        data=np.array(
            [
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, -1],
            ]
        )
    )

    for i in range(1000):
        valid_orient = m3d.Orientation.from_rotation_vector(np.random.rand(3))
        assert valid_orient.is_valid()

        invalid_orient = valid_orient * reflection
        assert not invalid_orient.is_valid()


def test_valid_rotation():
    t = m3d.Transform()
    assert t.is_valid()
    t.orient.rotate_xb(1)
    t.orient.rotate_yt(-1)
    assert t.is_valid()
    t._data[0, 0] = 2
    assert not t.is_valid()


def test_transform_from_corresponding_points():
    p1 = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 1.0],
            [1.0, 1.0, 1.0],
        ]
    )
    t = m3d.Transform.from_pose_vector(0.0, 1.0, 2.0, 0.1, 0.2, 0.3)
    p2 = t * p1
    t2 = m3d.Transform.from_corresponding_points(p1, p2)
    assert t == t2


def test_rotated():
    I = m3d.Transform()
    assert I.rotated_zb(np.pi).similar(I.rotated_zt(np.pi))

    X = m3d.Transform(vector=m3d.Vector(x=1.0))
    X_rot_b = X.rotated_zb(np.pi)
    X_rot_t = X.rotated_zt(np.pi)

    assert X_rot_t.pos.similar(X.pos)
    assert not X_rot_b.pos.similar(X.pos)
    assert not X_rot_t.orient.similar(X.orient)
    assert X_rot_b.orient.similar(X_rot_t.orient)



from m3d.transform import Transform
from m3d.vector import Vector
from m3d.orientation import Orientation
from m3d.common import float_eps
from m3d.quaternion import Quaternion
from m3d.dual_quaternion import DualQuaternion

__all__ = ["Transform", "Vector", "Orientation", "float_eps", "Quaternion", "DualQuaternion"]



from setuptools import setup, find_packages

setup(
    name="m3d",
    version="0.5.13",
    description="Python3 3D transformation library with object oriented API and MIT licensed",
    author="Olivier Roulet-Dubonnet",
    author_email="olivier.roulet@gmail.com",
    url="https://gitlab.com/kurant-open/m3d",
    package_data={
        "": ["py.typed"],
    },
    packages=find_packages(),
    provides=["m3d"],
    license="MIT",
    install_requires=[
        "numpy",
    ],
    extras_require={
        "orientation-mean": ["scipy"],
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
